"""Example of count data sampled from negative-binomial distribution
"""
import numpy as np
from matplotlib import pyplot as plt
from scipy import stats
from sklearn.model_selection import train_test_split

from xgboost_distribution import XGBDistribution


def generate_count_data(n_samples=10_000):
    X = np.random.uniform(-2, 0, n_samples)
    n = 66 * np.abs(np.cos(X))
    p = 0.5 * np.abs(np.cos(X / 3))

    y = np.random.negative_binomial(n=n, p=p, size=n_samples)
    return X[..., np.newaxis], y


def predict_distribution(model, X, y):
    """Predict a distribution for a given X, and evaluate over y"""

    distribution_func = {
        "normal": getattr(stats, "norm").pdf,
        "laplace": getattr(stats, "laplace").pdf,
        "poisson": getattr(stats, "poisson").pmf,
        "negative-binomial": getattr(stats, "nbinom").pmf,
    }
    preds = model.predict(X[..., np.newaxis])

    dists = np.zeros(shape=(len(X), len(y)))
    for ii, x in enumerate(X):
        params = {field: param[ii] for (field, param) in zip(preds._fields, preds)}
        dists[ii] = distribution_func[model.distribution](y, **params)

    return dists


def create_distribution_heatmap(
    model, x_range=(-2, 0), x_steps=100, y_range=(0, 100), normalize=True
):
    xx = np.linspace(x_range[0], x_range[1], x_steps)
    yy = np.linspace(y_range[0], y_range[1], y_range[1] - y_range[0] + 1)
    ym, xm = np.meshgrid(xx, yy)

    z = predict_distribution(model, xx, yy)

    if normalize:
        z = z / z.max(axis=0)

    return ym, xm, z.transpose()


def main():
    random_state = 10
    np.random.seed(random_state)

    X, y = generate_count_data(n_samples=10_000)
    X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=random_state)

    model = XGBDistribution(
        distribution="negative-binomial",  # try changing the distribution here
        natural_gradient=True,
        max_depth=3,
        n_estimators=500,
    )
    model.fit(
        X_train,
        y_train,
        eval_set=[(X_test, y_test)],
        early_stopping_rounds=10,
        verbose=False,
    )

    xm, ym, z = create_distribution_heatmap(model)

    fig, ax = plt.subplots(figsize=(9, 6))
    ax.pcolormesh(
        xm, ym, z, cmap="Oranges", vmin=0, vmax=1.608, alpha=1.0, shading="auto"
    )
    ax.scatter(X_test, y_test, s=0.75, alpha=0.25, c="k", label="data")
    plt.show()


if __name__ == "__main__":
    main()



"""Laplace distribution
"""
import numpy as np
from scipy.stats import cauchy, laplace

from xgboost_distribution.distributions.base import BaseDistribution


class Laplace(BaseDistribution):
    """Laplace distribution with log scoring

    Definition:

        f(x) = 1/2 * exp( -| (x - loc) / scale | ) / scale

    We reparameterize:
        loc   -> loc = a
        scale -> log(scale) = b to ensure scale >= 0.

    Thus:

        f(x) = 1/2 * exp( -| (x - a) / e^b | ) / e^b

    We compute the gradients:

        d/da -log[f(x)] = (a - x) / (scale * | a-x | )
        d/db -log[f(x)] = 1 - | a-x | / scale

    To second order:

        d2/da2 -log[f(x)] = 2 δ(x-a) / scale
        d2/db2 -log[f(x)] = | a-x | / scale

    The Fisher information is:

        I(loc) = 1 / scale^2
        I(scale) = 1 / scale^2

    which needs to be expressed in reparameterized form:

        1 / scale^2 = I_r(loc) (d/d(loc) loc) ^2
                    = I_r(loc)

        1 / scale^2 = I_r(scale) ( d/d(scale) log(scale) )^2
                    = I_r(scale) ( 1/ scale )^2

    Hence:

        I_r(loc) = 1 / scale^2
        I_r(scale) = 1

    """

    @property
    def params(self):
        return ("loc", "scale")

    def gradient_and_hessian(self, y, params, natural_gradient=True):
        """Gradient and diagonal hessian"""

        loc, log_scale = self._split_params(params)
        scale = np.exp(log_scale)

        grad = np.zeros(shape=(len(y), 2))
        grad[:, 0] = np.sign(loc - y) / scale
        grad[:, 1] = 1 - np.abs(loc - y) / scale

        if natural_gradient:
            fisher_matrix = np.zeros(shape=(len(y), 2, 2))
            fisher_matrix[:, 0, 0] = 1 / scale ** 2
            fisher_matrix[:, 1, 1] = 1

            grad = np.linalg.solve(fisher_matrix, grad)
            hess = np.ones(shape=(len(y), 2))  # we set the hessian constant
        else:
            hess = np.zeros(shape=(len(y), 2))  # diagonal elements only
            # Note: Delta functions won't work well, hence we approximate with cauchy
            hess[:, 0] = 2 * cauchy.pdf(y, loc, scale) / scale
            hess[:, 1] = 1 - grad[:, 1]

        return grad, hess

    def loss(self, y, params):
        loc, scale = self.predict(params)
        return "LaplaceError", -laplace.logpdf(y, loc=loc, scale=scale).mean()

    def predict(self, params):
        loc, log_scale = self._split_params(params)
        scale = np.exp(log_scale)
        return self.Predictions(loc=loc, scale=scale)

    def starting_params(self, y):
        return np.mean(y), np.log(np.std(y))

    def _split_params(self, params):
        """Return loc and log_scale from params"""
        return params[:, 0], params[:, 1]



"""XGBDistribution model
"""
from typing import Callable

import numpy as np
from sklearn.base import RegressorMixin
from sklearn.utils.validation import check_is_fitted
from xgboost import config_context
from xgboost.core import DMatrix
from xgboost.sklearn import XGBModel, _wrap_evaluation_matrices, xgboost_model_doc
from xgboost.training import train

from xgboost_distribution.distributions import get_distribution, get_distribution_doc


@xgboost_model_doc(
    "Implementation of XGBoost to estimate distributions (in scikit-learn API).",
    ["estimators", "model"],
    extra_parameters=get_distribution_doc()
    + """
    natural_gradient : bool, default=True
        Whether or not natural gradients should be used.""",
)
class XGBDistribution(XGBModel, RegressorMixin):
    def __init__(
        self, distribution=None, natural_gradient=True, objective=None, **kwargs
    ):
        self.distribution = distribution or "normal"
        self.natural_gradient = natural_gradient

        if objective is not None:
            raise ValueError(
                "Please do not set objective directly! Use the `distribution` kwarg"
            )

        super().__init__(objective=None, **kwargs)

    def fit(
        self,
        X,
        y,
        *,
        sample_weight=None,
        eval_set=None,
        early_stopping_rounds=None,
        verbose=False,
        xgb_model=None,
        sample_weight_eval_set=None,
        feature_weights=None,
        callbacks=None,
    ):
        """Fit gradient boosting distribution model.

        Note that calling ``fit()`` multiple times will cause the model object to be
        re-fit from scratch. To resume training from a previous checkpoint, explicitly
        pass ``xgb_model`` argument.

        Parameters
        ----------
        X : array_like
            Feature matrix
        y : array_like
            Labels
        sample_weight : array_like
            instance weights
        eval_set : list
            A list of (X, y) tuple pairs to use as validation sets, for which
            metrics will be computed.
            Validation metrics will help us track the performance of the model.
        early_stopping_rounds : int
            Activates early stopping. Validation metric needs to improve at least once
            in every **early_stopping_rounds** round(s) to continue training.
            Requires at least one item in **eval_set**.

            The method returns the model from the last iteration (not the best one).
            If there's more than one item in **eval_set**, the last entry will be used
            for early stopping.

            If early stopping occurs, the model will have three additional fields:
            ``clf.best_score``, ``clf.best_iteration``.
        verbose : bool
            If `verbose` and an evaluation set is used, writes the evaluation metric
            measured on the validation set to stderr.
        xgb_model : `xgboost.core.Booster`, `xgboost.sklearn.XGBModel`
            file name of stored XGBoost model or 'Booster' instance XGBoost model to be
            loaded before training (allows training continuation).
        sample_weight_eval_set : array_like
            A list of the form [L_1, L_2, ..., L_n], where each L_i is an array like
            object storing instance weights for the i-th validation set.
        feature_weights : array_like
            Weight for each feature, defines the probability of each feature being
            selected when colsample is being used.  All values must be greater than 0,
            otherwise a `ValueError` is thrown.  Only available for `hist`, `gpu_hist`
            and `exact` tree methods.
        callbacks : list
            List of callback functions that are applied at end of each iteration.
            It is possible to use predefined callbacks by using :ref:`callback_api`.
            Example:

            .. code-block:: python

                callbacks = [xgb.callback.EarlyStopping(rounds=early_stopping_rounds,
                                                        save_best=True)]

        """
        self._distribution = get_distribution(self.distribution)
        self._distribution.check_target(y)

        params = self.get_xgb_params()
        params["objective"] = None
        params["disable_default_eval_metric"] = True
        params["num_class"] = len(self._distribution.params)

        # we set `base_score` to zero and instead use base_margin in dmatrices
        # -> this allows different starting values for each distribution parameter
        params["base_score"] = 0.0
        self._starting_params = self._distribution.starting_params(y)

        base_margin = self._get_base_margin(len(y))
        if eval_set is not None:
            base_margin_eval_set = [
                self._get_base_margin(len(evals[1])) for evals in eval_set
            ]
        else:
            base_margin_eval_set = None

        train_dmatrix, evals = _wrap_evaluation_matrices(
            missing=self.missing,
            X=X,
            y=y,
            group=None,
            qid=None,
            sample_weight=sample_weight,
            base_margin=base_margin,
            feature_weights=feature_weights,
            eval_set=eval_set,
            sample_weight_eval_set=sample_weight_eval_set,
            base_margin_eval_set=base_margin_eval_set,
            eval_group=None,
            eval_qid=None,
            create_dmatrix=lambda **kwargs: DMatrix(nthread=self.n_jobs, **kwargs),
            label_transform=lambda x: x,
        )

        evals_result = {}
        model, _, params = self._configure_fit(xgb_model, None, params)

        # Suppress warnings from unexpected distribution & natural_gradient params
        with config_context(verbosity=0):
            self._Booster = train(
                params,
                train_dmatrix,
                num_boost_round=self.get_num_boosting_rounds(),
                evals=evals,
                early_stopping_rounds=early_stopping_rounds,
                evals_result=evals_result,
                obj=self._objective_func(),
                feval=self._evaluation_func(),
                verbose_eval=verbose,
                xgb_model=model,
                callbacks=callbacks,
            )

        self._set_evaluation_result(evals_result)
        self.objective = f"distribution:{self.distribution}"

        return self

    def predict(
        self,
        X,
        ntree_limit=None,
        validate_features=False,
        iteration_range=None,
    ):
        """Predict all params of distribution of each `X` example.

        Parameters
        ----------
        X : array_like
            Feature matrix.
        ntree_limit : int
            Deprecated, use `iteration_range` instead.
        validate_features : bool
            When this is True, validate that the Booster's and data's feature_names are
            identical.  Otherwise, it is assumed that the feature_names are the same.
        iteration_range :
            Specifies which layer of trees are used in prediction.  For example, if a
            random forest is trained with 100 rounds.  Specifying `iteration_range=(10,
            20)`, then only the forests built during [10, 20) (half open set) rounds are
            used in this prediction.

        Returns
        -------
        predictions : namedtuple
            A namedtuple of the distribution parameters. Each parameter is a
            numpy array of shape (n_samples,).
        """
        check_is_fitted(self, attributes=("_distribution", "_starting_params"))

        base_margin = self._get_base_margin(X.shape[0])

        params = super().predict(
            X=X,
            output_margin=True,
            ntree_limit=ntree_limit,
            validate_features=validate_features,
            base_margin=base_margin,
            iteration_range=iteration_range,
        )
        return self._distribution.predict(params)

    def save_model(self, fname) -> None:
        # self._distribution class cannot be saved by `super().save_model`, as it
        # attempts to call `json.dumps({"_distribution": self._distribution})`
        # Hence we delete, and then reinstantiate
        # (this is safe as distributions are by definition stateless)
        del self._distribution
        super().save_model(fname)
        self._distribution = get_distribution(self.distribution)

    def load_model(self, fname) -> None:
        super().load_model(fname)
        # See above: Currently need to reinstantiate distribution post loading
        self._distribution = get_distribution(self.distribution)

    def _objective_func(self) -> Callable:
        def obj(params: np.ndarray, data: DMatrix):
            y = data.get_label()
            grad, hess = self._distribution.gradient_and_hessian(
                y=y, params=params, natural_gradient=self.natural_gradient
            )
            return grad.flatten(), hess.flatten()

        return obj

    def _evaluation_func(self) -> Callable:
        def feval(params: np.ndarray, data: DMatrix):
            y = data.get_label()
            return self._distribution.loss(y=y, params=params)

        return feval

    def _get_base_margin(self, n_samples):
        return (
            np.ones(shape=(n_samples, 1)) * np.array(self._starting_params)
        ).flatten()



"""Test suite for XGBDistribution model
"""

import os
import pickle

import pytest

import numpy as np
from sklearn.exceptions import NotFittedError

from xgboost_distribution.distributions import AVAILABLE_DISTRIBUTIONS
from xgboost_distribution.model import XGBDistribution


def test_XGBDistribution_early_stopping_fit(small_train_test_data):
    """Integration test to ensure end-to-end functionality"""

    X_train, X_test, y_train, y_test = small_train_test_data
    model = XGBDistribution(distribution="normal", max_depth=3, n_estimators=500)
    model.fit(X_train, y_train, eval_set=[(X_test, y_test)], early_stopping_rounds=10)

    evals_result = model.evals_result()

    assert model.best_iteration == 6
    assert isinstance(evals_result, dict)


def test_XGBDistribution_early_stopping_predict(small_train_test_data):
    """Check that predict with early stopping uses correct ntrees"""
    X_train, X_test, y_train, y_test = small_train_test_data

    model = XGBDistribution(distribution="normal", max_depth=3, n_estimators=500)
    model.fit(X_train, y_train, eval_set=[(X_test, y_test)], early_stopping_rounds=10)

    mean, var = model.predict(X_test)
    mean_iter, var_iter = model.predict(
        X_test, iteration_range=(0, model.best_iteration + 1)
    )
    np.testing.assert_array_equal(mean, mean_iter)
    np.testing.assert_array_equal(var, var_iter)


def test_objective_overwrite_by_repeated_fit(small_X_y_data):
    """In the fit step, we set the objective to distribution:normal (e.g.)

    This needs to be re-set at each fit, as it's not a standard xgboost objective.
    """
    X, y = small_X_y_data
    model = XGBDistribution()
    model.fit(X, y)
    model.fit(X, y)


def test_distribution_set_param(small_X_y_data):
    """Check that updating the distribution params works"""
    X, y = small_X_y_data

    model = XGBDistribution(distribution="abnormal")
    with pytest.raises(ValueError):
        model.fit(X, y)  # fails as distribution does not exist

    model.set_params(**{"distribution": "normal"})
    model.fit(X, y)


# -------------------------------------------------------------------------------------
# Failure modes
# -------------------------------------------------------------------------------------


def test_fit_1d_X_fit_fails(small_X_y_data):
    X, y = small_X_y_data
    X = X[..., 0]  # single 1d features

    model = XGBDistribution()
    with pytest.raises(ValueError):
        model.fit(X, y)


def test_predict_before_fit_fails(small_X_y_data):
    X, y = small_X_y_data

    model = XGBDistribution()
    with pytest.raises(NotFittedError):
        model.predict(X)


def test_setting_objective_in_init_fails():
    with pytest.raises(ValueError):
        XGBDistribution(objective="binary:logistic")


# -------------------------------------------------------------------------------------
#  Model internal tests
# -------------------------------------------------------------------------------------


@pytest.mark.parametrize(
    "distribution, expected_margin",
    [
        ("normal", np.array([0.5, np.log(0.5), 0.5, np.log(0.5)])),
        ("poisson", np.array([np.log(0.5), np.log(0.5)])),
    ],
)
def test_get_base_margin(distribution, expected_margin):
    """Test that base_margin are created as expected for each sample"""
    model = XGBDistribution(distribution=distribution)
    X = np.array([[1], [1]])
    y = np.array([1, 0])
    model.fit(X, y)

    margin = model._get_base_margin(n_samples=2)
    np.testing.assert_array_equal(margin, expected_margin)


@pytest.mark.parametrize(
    "distribution",
    list(AVAILABLE_DISTRIBUTIONS.keys()),
)
def test_objective_and_evaluation_funcs_callable(distribution):
    model = XGBDistribution(distribution=distribution)
    assert callable(model._objective_func())
    assert callable(model._evaluation_func())


# -------------------------------------------------------------------------------------
#  Model IO tests
# -------------------------------------------------------------------------------------


def assert_model_equivalence(model_a, model_b, X):
    np.testing.assert_almost_equal(
        model_a._starting_params, model_b._starting_params, decimal=9
    )
    assert model_a._distribution.__class__ == model_b._distribution.__class__

    preds_a = model_a.predict(X)
    preds_b = model_b.predict(X)

    for param_a, param_b in zip(preds_a, preds_b):
        np.testing.assert_array_equal(param_a, param_b)


@pytest.mark.parametrize(
    "model_format",
    ["bst", "json"],
)
def test_XGBDistribution_save_and_load_model(small_X_y_data, model_format, tmpdir):
    X, y = small_X_y_data
    model = XGBDistribution(n_estimators=10)
    model.fit(X, y)

    model_path = os.path.join(tmpdir, f"model.{model_format}")
    model.save_model(model_path)

    saved_model = XGBDistribution()
    saved_model.load_model(model_path)

    assert_model_equivalence(model_a=model, model_b=saved_model, X=X)


def test_XGBDistribution_pickle_dump_and_load(small_X_y_data, tmpdir):
    X, y = small_X_y_data
    model = XGBDistribution(n_estimators=10)
    model.fit(X, y)

    model_path = os.path.join(tmpdir, "model.pkl")

    with open(model_path, "wb") as fd:
        pickle.dump(model, fd)

    with open(model_path, "rb") as fd:
        saved_model = pickle.load(fd)

    assert_model_equivalence(model_a=model, model_b=saved_model, X=X)



import pytest

import numpy as np
import pandas as pd

from xgboost_distribution.distributions import LogNormal


@pytest.fixture
def lognormal():
    return LogNormal()


def test_target_validation(lognormal):
    valid_target = np.array([0.5, 1, 4, 5, 10])
    lognormal.check_target(valid_target)


@pytest.mark.parametrize(
    "invalid_target",
    [np.array([0, 1.2]), pd.Series([-1.1, 0.4, 2.3])],
)
def test_target_validation_raises(lognormal, invalid_target):
    with pytest.raises(ValueError):
        lognormal.check_target(invalid_target)


@pytest.mark.parametrize(
    "y, params, natural_gradient, expected_grad",
    [
        (
            np.array([1, 1]),
            np.array([[np.log(1), 2], [1, 0]]),
            True,
            np.array([[0, 0.5], [1, 0]]),
        ),
        (
            np.array([1, 1]),
            np.array([[np.log(1), 2], [1, 0]]),
            False,
            np.array([[0, 1], [1, 0]]),
        ),
    ],
)
def test_gradient_calculation(lognormal, y, params, natural_gradient, expected_grad):
    grad, hess = lognormal.gradient_and_hessian(
        y, params, natural_gradient=natural_gradient
    )
    np.testing.assert_array_equal(grad, expected_grad)


def test_loss(lognormal):
    loss_name, loss_value = lognormal.loss(
        # fmt: off
        y=np.array([0, ]),
        params=np.array([[1, 0], ]),
    )
    assert loss_name == "LogNormalError"
    assert loss_value == np.inf
